/**
 * @version 3.5 2004-04-12
 * @author Gonzalo Ojeda 9821018-0
 */
 
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;
import java.io.*;

/**
 * La clase <b>FigPanel</b> corresponde a la zona de dibujo, donde se ubican 
 * las figuras y se realizan las operaciones sobre estas mismas. 
 */
public class FigPanel extends JPanel
{
   /**
    * Constructor de la clase <b>FigPanel</b>, define los ActionListener para
    * los botones de Figuras y de Operaciones, adems realiza todas las 
    * operaciones, tanto la creacin
    * de las figuras como las operaciones sobre ellas.
    * @param rectButton Boton que indica que la proxima accin es dibujar 
    * <i>Rectngulos</i>
    * @param elipButton Boton que indica que la proxima accin es dibujar 
    * <i>Elpses</i>
    * @param linButton  Boton que indica que la proxima accin es dibujar 
    * <i>Lneas</i>
    * @param rotButton  Boton que indica que la proxima accin es <i>Rotar</i> 
    * el dibujo
    * @param escButton  Boton que indica que la proxima accin es <i>Escalar</i>
    * el dibujo
    * @param trasButton Boton que indica que la proxima accin es 
    * <i>Trasladar</i> el dibujo
    */
   public FigPanel (JButton rectButton, JButton elipButton, JButton linButton,
      JButton rotButton, JButton escButton, JButton trasButton)
   {
      // setea el color de la zona de dibujo, en este caso, blanco
      setBackground(Color.WHITE);
      // arreglo de largo variable para almacenar todas las figuras a crear
      shapes = new ArrayList();
      formas = new ArrayList();     
      // la figura actual, se la inicializa vaca.
      current = null;
      // necesaria para las operaciones que funcionan con mas de un click
      contador = 0;
      // puede ser un punto cualquiera, la idea es tener un valor para comparar
      // y comprobar que se han hecho 2 clicks del mouse y no 1
      nullPoint = new Point2D.Double (0, 0);
      // se crean listener para todas las operaciones del mouse.
      addMouseListener(new MouseHandler());
      addMouseMotionListener (new MouseMotionHandler());
      // ActionListener para responder a los botones de Figura y Operacin,
      // almacena la accin realizada en el string figura, y limpia el contador
      ActionListener figListener = new ActionListener ()
      {
         public void actionPerformed (ActionEvent e)
         {
            figura = e.getActionCommand();
            contador = 0;
         }
      };
      // se agrega el mismo action listener a los 6 botones, pues las 6 acciones
      // son excluyentes entre s.
      rectButton.addActionListener(figListener);
      elipButton.addActionListener(figListener);
      linButton.addActionListener(figListener);
      rotButton.addActionListener(figListener);
      escButton.addActionListener(figListener);
      trasButton.addActionListener(figListener);
   }
   
   /**
    * Mtodo que es invocada por la mquina virtual cuando se realiza una 
    * peticin de repintado (<i>repaint()</i>), invoca al mtodo paintComponent
    * de la clase superior. Este mismo mtodo realiza la rotacin, escalamiento,
    * o traslacin del dibujo completo en caso de ser solicitado (click en el 
    * respectivo botn) sin embargo, estos 3 cambios son solo temporales hasta
    * que no se inserte una  nueva figura, pues los cambios se hacen sobre el
    * objeto grfico <i>g2</i> y no sobre el arreglo de las figuras, por la
    * imposibilidad de guardar figuras rotadas.
    * @param g    Objeto entregado por la mquina virtual que permite dibujar
    *             las figuras previo casteo a un objeto de tipo Graphics2D
    */
   public void paintComponent(Graphics g)
   {
      super.paintComponent(g);
      Graphics2D g2 = (Graphics2D)g;

      // slo si est marcado el botn Rotar, se realiza esta accin. 
      if ("Rotar".equals(figura))
         g2.rotate(tetha, iniRot.getX(), iniRot.getY());
            
      // slo si est marcado el botn Escalar, se realiza esta accin.
      else if ("Escalar".equals(figura))
         g2.scale(zoom, zoom);

      // slo si est marcado el botn Trasladar, se realiza esta accin.     
      else if ("Trasladar".equals(figura))
         g2.translate(dx, dy);

      // dibujar todas las formas contenidas en el arreglo de largo variable
      for (int i = 0; i < shapes.size(); i++)
         g2.draw((Shape)shapes.get(i));
   }
   
   /**
    * Mtodo que agrega una nueva figura al arreglo de largo variable, segn sea
    * la opcin seleccionada anteriormente.
    * El Mtodo trabaja sobre 2 puntos, al realizar el primer click se definen
    * estos 2 puntos al mismo tiempo, y a medida que se mueve el mouse con el
    * botn presionado se va variando el segundo punto.
    * Las formas creadas son <i>Rectangle2D.Double</i> para Rectngulos, 
    * <i>Ellipse2D.Double</i> para las Elpses, y <i>Line2D.Double</i> para las 
    * lneas, estas 3 clases heredan de la Interfase <i>Shape</i>, por lo que se
    * puede usar un objeto de este tipo para trabajar con los 3 tipos de figura.
    */
   public void add ()
   {
      double a = inicio.getX();
      double b = inicio.getY();
      double x = last.getX();
      double y = last.getY();

      // Crea Rectngulos, y luego los agrega al arreglo, se redibuja todo.
      if ("Rectangulo".equals(figura))
      {
//       current = new Rectangle2D.Double (a, b, x - a, y - b);
         actual = new Rectangulo (inicio, last);
         formas.add(actual);
         current = (Rectangle2D)actual.setForma();         
         shapes.add(current);
         repaint();
      }
      // Crear Elpses, y luego los agrega al arreglo, se redibuja todo
      else if ("Elipse".equals(figura))
      {
//       current = new Ellipse2D.Double(a, b, x - a, y - a);
         actual = new Elipse (inicio, last);
         formas.add(actual);
         current = (Ellipse2D)actual.setForma();
         shapes.add(current);
         repaint();
      }
      // Crear Lneas, luego de crearlo y agregarlo al arreglo, se redibuja todo
      else if ("Linea".equals(figura))
      {
//       current = new Line2D.Double(inicio, last);
         actual = new Linea (inicio, last);
         formas.add(actual);
         current = (Line2D)actual.setForma();
         shapes.add(current);
         repaint();
      }
   }
   
   /**
    * Mtodo que limpia la zona de dibujo, borra todos los elementos del arreglo
    * de todo y actualiza la zona de dibujo con ningn elemento.
    * Este mtodo es invocado al presionar el botn limpiar, y es llamado desde
    * el otro archivo por la clase <i>EditorFrame</i>,tambien limpia el contador
    */
   public void clearShapes()
   {
      shapes.clear();
      formas.clear();
      contador = 0;
      repaint();
   }

   /**
    * Mtodo que rescata las figuras desde el archivo ".eda", las copia al 
    * arreglo de Shapes y las dibuja.
    */
   public void setShapes()
   {
      // reconstruir el arreglo de shapes al abrir archivo
      for (int i = 0; i < formas.size(); i++)
      {
         // si es un rectangulo, crear un rectangulo con los puntos
         if ((Formas)formas.get(i) instanceof Rectangulo)
         {
            actual = (Rectangulo)formas.get(i);
            double a = actual.getP1().getX();
            double b = actual.getP1().getY();
            double x = actual.getP2().getX();
            double y = actual.getP2().getY();
            current = new Rectangle2D.Double (a, b, x - a, y - b);
            shapes.add(current);
         }
         // si es una elipse, crear una elipse con los puntos
         else if ((Formas)formas.get(i) instanceof Elipse)
         {
            actual = (Elipse)formas.get(i);
            double a = actual.getP1().getX();
            double b = actual.getP1().getY();
            double x = actual.getP2().getX();
            double y = actual.getP2().getY();
            current = new Ellipse2D.Double (a, b, x - a, y - b);
            shapes.add(current);
         }
         // si es una lnea, crear una lnea con los puntos
         else if ((Formas)formas.get(i) instanceof Linea)
         {
            actual = (Linea)formas.get(i);
            current = new Line2D.Double (actual.getP1(), actual.getP2());
            shapes.add(current);
         }
      }
      repaint();        // redibuja todas las figuras
   }

   /**
    * Mtodo que guarda las figuras en los archivos
    * @param name nombre de archivo donde se guardarn las figuras
    * @throws IOException Si orcurre un error de Entrada/Salida al escribir
    * al archivo
    */
   public void save(String name) throws IOException
   {
      final int EXTENSION = 4;   //largo de una extension (ej: .eda)
      // si se pasa la extensin .eda al archivo, no se agrega denuevo y se
      // saca para el archivo .fig
      if (name.toLowerCase().endsWith(".eda"))
      {
         fileName = name;
         String s = name.substring(0, name.length() - EXTENSION);
         fileName2 = new String (s + ".fig");
      }
      // si se pasa la extensin .fig al archivo, no se agrega denuevo y se
      // saca para el archivo .eda
      else if (name.toLowerCase().endsWith(".fig"))
      {
         String s = name.substring(0, name.length() - EXTENSION);
         fileName = new String(s + ".eda");
         fileName2 = name;
      }
      // si no se pasa extensin, se agregan normalmente
      else
      {
         fileName = new String (name + ".eda");
         fileName2 = new String (name + ".fig");
      }
      // se escribe el archivo .fig
      FigWrite fig = new FigWrite (fileName2, formas);
      // si el archivo .eda existe se lo borra antes, sino agrega al final
      File arch = new File(fileName);
      arch.delete();
      // escribe el ArrayList de formas al archivo .eda
      FileOutputStream ostream = new FileOutputStream(fileName);
      ObjectOutputStream p = new ObjectOutputStream(ostream);
      p.writeObject(formas);
      p.flush();
      ostream.close();
   }

   /**
    * Mtodo que lee las figuras desde el archivo ".eda".
    * @param name nombre de archivo donde se guardarn las figuras
    * @throws IOException Si orcurre un error de Entrada/Salida al escribir
    * al archivo
    * @throws ClassNotFoundException si se encuentra un objeto de una clase no
    * serializable
    */
   public void read(String name) throws IOException, ClassNotFoundException
   {
      clearShapes();    // borra los ArrayList y limpia la zona de dibujo
      // lee el ArrayList desde el archivo
      FileInputStream istream = new FileInputStream(name);
      ObjectInputStream p = new ObjectInputStream(istream);
      formas = (ArrayList)p.readObject();
      istream.close();     // cierra el InputStram
      // invoca a la funcin que reconstruye las figuras
      setShapes();
   }
   
   /** 
    * Clase interna que maneja los clicks del mouse, extiende la clase 
    * MouseAdapter por lo tanto solo es necesario definir las funciones que se 
    * van a utilizar del mouse, en este caso solo mousePressed.
    */
   private class MouseHandler extends MouseAdapter
   {
      /**
       * Mtodo llamado al presionar el mouse, al presionar el mouse se agregan
       * el punto inicial y final de la figura (todas las formas a dibujar 
       * requieren de una u otra forma 2 puntos para ser dibujadas).
       * En el caso de tratarse de una figura, se definen los 2 puntos altiro, y
       * se llama al mtodo <i>add()</i> para crear la figura.
       * Para el caso de Rotar se define un punto para el primer click, y otro 
       * para el segundo para despues en base a el movimiento del mouse calcular
       * el angulo de rotacin. Para el caso de Escalar se define un punto al 
       * primer click y el segundo punto al siguiente click, y segun el 
       * movimiento del mouse se hace "zoom".
       */
      public void mousePressed (MouseEvent event)
      {
         if (("Rectangulo".equals(figura)) || ("Elipse".equals(figura)) 
            || ("Linea".equals(figura)))
         {
            inicio = event.getPoint();
            last = event.getPoint();
            add();
         }
         else if ("Rotar".equals(figura))
         {
            if (contador == 0)
            {
               iniRot = event.getPoint();
               // al hacer el primer click se hace esto porque hay problemas se
               // intenta rotar al presionar el boton por primer vez y arrastrar
               rotPoint = nullPoint;
               contador = 1;
            }
            else
            {
               rotPoint = event.getPoint();
               contador--;
            }
         }
         else if ("Escalar".equals(figura))
         {
            if (contador == 0) 
            {
               trasPoint = event.getPoint();
               contador = 1;
               // para impedir que solo se presione una vez el mouse
               // para escalar (se necesitan 2 clicks
               iniDistancia = 0;
            }
            else
            {
               last = event.getPoint();
               contador--;
               // Se calcula la distancia inicial si ya se tienen los 2 puntos
               iniDistancia = trasPoint.distance(last);
            }
         }
         else if ("Trasladar".equals(figura))
         {
            trasPoint = event.getPoint();
         }
      }  
      
      public void mouseReleased(MouseEvent event)
      {
         if ("Trasladar".equals(figura))
         {
            dx1 = dx;
            dy1 = dy;
         }

         else if (("Rectangulo".equals(figura)) || ("Elipse".equals(figura)) 
            || ("Linea".equals(figura)))
         {
            int i = formas.size();
            actual = (Formas)formas.get(formas.size() - 1);
            actual.setP2(last);
         }
      }

   }  
   
   /**
    * Clase Interna para manejar los movimientos del mouse, en este caso el 
    * mtodo <i>mouseMoved</i> que corresponde a movimientos del mouse sin 
    * presionar el botn no es utilizado, as que no se implementa.
    * El mtodo <i>mouseDragged</i> se utiliza para actualizar las figuras, y
    * realizar  las operaciones mientras se mueve el mouse.
    * Lo primero que se hace es obtener el punto donde est actualmente el mouse
    * y  guardarlo.
    * Para la creacin de Rectngulos y Elpses se realiza un 
    * setFrameFromDiagonal, para actualizar la figura a medida que se mueve el 
    * mouse. Para lneas se usa setLine. Para las operaciones se obtiene el 
    * punto final, se realizan algunos calculos en caso de ser necesarios, y se 
    * llama al mtodo <i>repaint()</i> para actualizar el dibujo.
    */
   private class MouseMotionHandler implements MouseMotionListener
   {
      public void mouseMoved (MouseEvent event) 
         { };
      
      public void mouseDragged (MouseEvent event)
      {
         last = event.getPoint();
         
         if ("Rectangulo".equals(figura))
         {        
            Rectangle2D rect = (Rectangle2D)current;
            rect.setFrameFromDiagonal(inicio, last);
            repaint();
         }
         else if ("Elipse".equals(figura))
         {
            Ellipse2D elip = (Ellipse2D)current;
            elip.setFrameFromDiagonal(inicio, last);
            repaint();
         }
         else if ("Linea".equals(figura))
         {
            Line2D lin = (Line2D)current;
            lin.setLine (inicio, last);
            repaint();
         }
         else if ("Trasladar".equals(figura))
         {
            dx = event.getX() - trasPoint.getX();
            dy = event.getY() - trasPoint.getY();
            dx = dx1 + dx;
            dy = dy1 + dy;
            repaint();
         }
         else if ("Escalar".equals(figura))
         {
            // no escala si solo se ha presionado una vez el mouse
            if (iniDistancia == 0)
               return;
            // se calcula la distancia desde el punto inicial al punto actual
            finDistancia = trasPoint.distance(last);
            // calcular el factor para hacer zoom dividiendo la distancia actual
            // por la distancia inicial entre los 2 puntos.
            zoom = finDistancia / iniDistancia;
            repaint();
         }
         else if ("Rotar".equals(figura))
         {
            // si solo ha habido un click, no rotar.
            if (rotPoint.equals(nullPoint))
               return;
            // el clculo del ngulo se realiza por medio del teorema del coseno
            // la variable temp se uso para calcular b^2 + c^2 - a^2
            double temp = iniRot.distanceSq(last) + iniRot.distanceSq(rotPoint)
               - rotPoint.distanceSq(last);
            tetha = Math.acos(temp / (2 * iniRot.distance(last) 
               * iniRot.distance(rotPoint)));
            // El clculo anterior solo devuelve valores positivos, para obtener
            // ngulos negativos cuando corresponda se realiza lo siguiente.
            if (last.getY() < rotPoint.getY())
               tetha = -tetha;
            repaint();
         }
      }  
   }
   
   // Arreglo de largo variable que almacena las figuras como Shapes y Formas
   private ArrayList shapes, formas;
   private Shape current;     // figura actual
   private Formas actual;
   private String figura;     // almacena que se esta realizando segn botones
   private int contador;
   private double tetha;      // angulo para rotaciones.
   private double iniDistancia, finDistancia, zoom;   // para las operaciones.
   private Point2D inicio;     // punto inicial de las figuras
   private Point2D  last;      // punto final de las figuras
   private Point2D trasPoint, rotPoint, iniRot; // puntos para las operaciones
   private Point2D nullPoint; // punto para que no haya errores en rotar
   private double dx, dy, dx1 , dy1;      // deltas para el trasladar
   private String fileName, fileName2;    // nombre del archivo a abrir
  }
